//
// GameX
// Ship Demo - Ship Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include <math.h>

#include "shipdemox-ship.hpp"

extern Photons photons; // so we can use the global Photons object defined in shipdemox-main.cpp

// Ship - Initialize
// Initialize all ship images and ship's target image
void Ship::Initialize (int num)
{
	ship_num = num;					// Initialize ship number													
	curr_img = SHIP_IMG_FLAT;		// Set current ship image
	refire = 0;
									// Set initial ship position and angle:
	if (ship_num==0) 	{x = (float)GameX.GetWidth()*1/3; y = (float)GameX.GetHeight()*1/2; ang =   0;}
	else				{x = (float)GameX.GetWidth()*2/3; y = (float)GameX.GetHeight()*1/2; ang = 180;}

									// Load Sound(s):
	photon_snd.LoadWav ("Sounds\\torpedo.wav");		
									// Load Images:
	ships.Load ("Images\\ships.gif", "Images\\ships_alpha.gif"); // Load the sprite sheet image containing the ships, and a transparency layer from a different image file
	ships.SwapColors((ColorSwapType)ship_num); // change to ship's color

	target.Load ("Images\\target.tif", true); // Load the ship's target image, with transparency from the same image file
	target.SwapColors((ColorSwapType)ship_num); // change to ship's color
}

// Ship - Update
// Get user input and move the ship
void Ship::Update (void)
{
	bool pressedLeft, pressedRight, pressedUp, pressedDown, pressedFire; // each is true if pressed, false if not

	if (ship_num==0) {		// Process user input for Blue Ship (num 0)...
		pressedLeft = GameX.IsKeyDown(KEY_LEFT);
		pressedRight = GameX.IsKeyDown(KEY_RIGHT);
		pressedUp = GameX.IsKeyDown(KEY_UP);
		pressedDown = GameX.IsKeyDown(KEY_DOWN);
		pressedFire = GameX.IsKeyDown(KEY_SPACE);
	} else {				// Process user input for Red Ship (num 1)...
		pressedLeft = GameX.IsKeyDown(KEY_A);
		pressedRight = GameX.IsKeyDown(KEY_D);
		pressedUp = GameX.IsKeyDown(KEY_W);
		pressedDown = GameX.IsKeyDown(KEY_S);
		pressedFire = GameX.IsKeyDown(KEY_X);
	}

	if (pressedLeft && pressedRight) pressedLeft = false, pressedRight = false; // Cancel out any
	if (pressedUp   && pressedDown ) pressedUp   = false, pressedDown  = false; // conflicting keypresses.

	if (pressedLeft) {							// Left Key Pressed
		dang += 1; curr_img++;					// Rotate ship left
		if (curr_img > SHIP_IMG_RIGHT) curr_img = SHIP_IMG_RIGHT;
	} else if (pressedRight) {					// Right Key Pressed
		dang -= 1; curr_img--;					// Rotate ship right
		if (curr_img < SHIP_IMG_LEFT) curr_img = SHIP_IMG_LEFT;
	} else {									// Otherwise..
		if (dang>0) dang *= 0.875f;				// Gradually stop the ship from
		if (dang<0) dang *= 0.875f;				// rotating
		if (curr_img < SHIP_IMG_FLAT) curr_img++;
		if (curr_img > SHIP_IMG_FLAT) curr_img--;
	}
	if (pressedUp) {							// Thrust Key Pressed, so
		vx += (float) cosf(ang*DEGtoRAD)*.4f;	// Increase ship velocity
		vy += (float)-sinf(ang*DEGtoRAD)*.4f;		
	} else if (pressedDown) {					// Reverse-Thrust Key Pressed, so
		vx -= (float) cosf(ang*DEGtoRAD)*.4f;	// Decrease ship velocity
		vy -= (float)-sinf(ang*DEGtoRAD)*.4f;
	} else {									// Otherwise, assume they mean to stop, so..
		vx *= 0.9f;								// Gradually (sort of) bring ship to a stop
		vy *= 0.9f;
	}
	if (pressedFire) {							// Fire Key Pressed
		if (refire > 3) {						// If refire counter is ready..
			float fire_x = vx + (float) cosf(ang*DEGtoRAD) * 20.0f; // Determine direction & velocity of 
			float fire_y = vy + (float)-sinf(ang*DEGtoRAD) * 20.0f; // the new 'fired' photon
						// Add a photon to the list of photons to be processed
			photons.AddPhoton (x, y, fire_x, fire_y);
			GameX.PlaySound (&photon_snd, PLAY_REWIND, 0.5f); // Play photon torpedo sound (at 50% volume since it's loud)
			refire = 0;							// Reset the refire readiness counter
		}
	} 	
												//*** Move the ship
	ang += dang;								// Adjust angle by change-in-angle	
	x += vx;									// Move ship position by ship's velocity
	y += vy;
	refire++;									// Increment the refire readiness counter

	if (x < 0) vx = -vx, x = 0;			// If ship went out of the view,
	if (y < 0) vy = -vy, y = 0;			// bring it back into view.
	if (x >= GameX.GetWidth())  vx = -vx, x = (float)GameX.GetWidth()-1;
	if (y >= GameX.GetHeight()) vy = -vy, y = (float)GameX.GetHeight()-1;
}

// Ship - CollideCheck
// Do ship collision detection with another ship
void Ship::CollideCheck (Ship* ship2)
{
	target_x = ship2->x - x;					// Calculate vector to 2nd ship
	target_y = ship2->y - y;
												// Determine distance to 2nd ship
	float dist = sqrtf(target_x*target_x + target_y*target_y);
	if(dist > 0) {
		target_x /= dist;						// Determine direction to 2nd ship
		target_y /= dist;						// (is normalized vector)
	}	
	if (dist < 40.0f*1.25f) {					// If distance to 2nd ship is less than a certain number of pixels..
		float speed = sqrtf(vx*vx + vy*vy);		// Calculate speed of this ship
		float speed2 = sqrtf(ship2->vx*ship2->vx + ship2->vy*ship2->vy);	// Calculate speed of other ship

		float vx2 = -target_x * (speed+speed2)*0.5f/2;	// Calculate 'rebound' vector to throw ships in opposite directions 
		float vy2 = -target_y * (speed+speed2)*0.5f/2;	// (with speed equal to the sum of speeds of the two ships)			

		vx += vx2;								// Add rebound to this ship's velocity
		vy += vy2;
		ship2->vx -= vx2;						// Add rebound to the other ship's velocity
		ship2->vy -= vy2;
	}
	target_x *= 100;					// Move target 100 pixels away from ship
	target_y *= 100;					// (so that we can draw it)
	ship2->target_x = -target_x; // Set the other ship to target this one
	ship2->target_y = -target_y;
}

// Ship - Draw
// Draw the ship's shadow, target, and the ship itself
void Ship::Draw (void)
{
	int left = (curr_img % 5) * (80 + 1); // pixel x-value of sprite start in source "ships" image
	int top = (curr_img / 5) * (80 + 1); // pixel y-value of sprite start in source "ships" image

	// Draw the ship's shadow (by drawing the ship sprite totally black and halfway translucent)
	GameX.SetDrawShading (ColorX(0,0,0,128)); // (0,0,0,128) corresponds to (red,green,blue,opacity), each is out of 255
	GameX.SetDrawPart (left, top, left+80, top+80); // the individual ship sprite is 80x80 pixels
	GameX.SetDrawMode (DRAWOP_CENTERED);
	GameX.DrawImage (&ships, ((int)x+6), ((int)y+6), ang-90, 1.25f); // the +4 offsets the shadow from the ship

	// Draw the ship itself, note we draw this AFTER the shadow so it shows up on top of the shadow
	GameX.SetDrawPart (left, top, left+80, top+80);
	GameX.SetDrawMode (DRAWOP_CENTERED);
	GameX.DrawImage (&ships, (int)x, (int)y, ang-90, 1.25f); // (ang-90) is rotation angle, 1.25 means slightly enlarged by scaling

	// Draw ship's "target" image
	GameX.SetDrawMode (DRAWOP_CENTERED);
	GameX.DrawImage (&target, (int)(x+target_x), (int)(y+target_y));
}