//
// GameX - WindowsDX Class Code
// NOTE: Not all WindowsDX functions are implemented in this file
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


// KNOWN ISSUES:

// Certain video cards (old RAGE cards)
// don't draw alpha images using their shading's transparency values
// (i.e. drawing a normal alpha image at 50% transparency results in it being drawn at 100%)
// This is because these video cards don't have support for more than 1 processing layer
// but perhaps a workaround can be implemented somehow?

#include "gamex-win-dx.hpp"




bool WindowsDX::LockBackBuffer (void)
{
	if(drawLocked) return true;
	
	// since this means a switch from D3D to DD drawing,
	// we have to interrupt batch rendering here
	if(draw_render_batch_open)
		EndRenderBatch();

	DDSURFACEDESC2		ddsd;

	ddsd.dwSize = sizeof(ddsd) ;

	HRESULT status = dd_back_buf->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);	// lock drawing surface
	if(FAILED(status)) {
		DirectDrawError(status);
		return false;
	}

	win_screen = (char *) ddsd.lpSurface;

	drawLocked = true;

	return true;
}

void WindowsDX::Begin3DScene (void)
{
	draw_sort_mode = SCENESORTZ|SCENESORTD|SCENESORT2;

	if(!d3d_zbuffer_enabled) // resort to depth sorting is Z-buffering isn't working
		if(draw_sort_mode != 0)
			draw_sort_mode = SCENESORTD;

	if(draw_sort_mode & SCENESORTD) {
		draw_3D_depth_sort = 1;
	}
	if(draw_sort_mode & SCENESORTZ) {
		d3d_device->SetRenderState(D3DRENDERSTATE_ZENABLE,TRUE);
		d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,TRUE);
	}
}

GameXDrawState * WindowsDX::ms_iter_merge(GameXDrawState * before, GameXDrawState * f1, int n1, GameXDrawState * f2, int n2)
{
	GameXDrawState * first = NULL;
	GameXDrawState * last = NULL;
	GameXDrawState * temp = NULL;
	first = last = ((f1->zcomp >= f2->zcomp) ? f1 : f2);
	for(int i = 0, j = 0 ; i < n1 || j < n2; ) {
		if (i < n1 && (j >= n2 || f1->zcomp >= f2->zcomp)) {
			temp = f1;
			f1 = f1->next;
			i++;
		} else {
			temp = f2;
			f2 = f2->next;
			j++;
		}
		last->next = temp;
		last = temp;
	}
	if(before == NULL)
		draw_state_head = first;
	else
		before->next = first;
	last->next = f2;
	return last;	   
}

void WindowsDX::End3DScene(void)
{
	if(draw_sort_mode & SCENESORTZ)
		d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,FALSE);

	if(draw_sort_mode & SCENESORTD) {


		for(int i = 1 ; i < draw_state_num ; i <<= 1) {
			int n1 = i;
			int n2 = i;
			GameXDrawState * before = NULL;
			for(int j = 0 ; j + n1 < draw_state_num ; j += i<<1) {
				GameXDrawState *f1, *f2;
				f1 = f2 = (before == NULL) ? draw_state_head : before->next;
				for(int k = 0 ; k < n1 ; k++) 
					f2 = f2->next;
				n2 = min(n2, draw_state_num-j-n1);	   
				before = ms_iter_merge(before, f1, n1, f2, n2);
			}
		}

		
		draw_3D_depth_sort = 2;

		GameXDrawState * initial = draw_state;
		draw_state = draw_state_head;
		while(draw_state != NULL) {
			if(draw_state->src != VIEWPORT && draw_state->src != NULL
			&& draw_state->flags & DRAWOP_ALPHAINTERSECT && draw_state->flags & DRAW_PLAIN
			&& draw_state->src->m_usage & LOAD_ALPHA
			&& draw_sort_mode & SCENESORTZ && draw_state->dst == VIEWPORT) {
				// if it meets certain conditions and DRAWOP_ALPHAINTERSECT is enabled,
				// draw it twice to allow the alpha image to intersect in 3D:
				d3d_device->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
				d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,TRUE);
				bool keep = (draw_state->flags & DRAWOP_KEEPSTATES) ? true : false;
				if(!keep) draw_state->flags |= DRAWOP_KEEPSTATES;
				InternalDrawMaster(0);
				d3d_device->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
				d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,FALSE);
				if(!keep) draw_state->flags ^= DRAWOP_KEEPSTATES;
				InternalDrawMaster(0);
			} else // otherwise, draw once as normal
				InternalDrawMaster(0);
			if(draw_state != NULL) {
				GameXDrawState * next = draw_state->next;
				if(draw_state == draw_state_head) draw_state_head = NULL; // causes hard-to-track pointer errors
				if(draw_state == draw_state_tail) draw_state_tail = NULL; // if we forget to reset head+tail to NULL when deleting
				if(draw_state != initial)
					delete draw_state;
				draw_state_num--;
				draw_state = NULL;
				draw_state = next;
			}
		}

		draw_state = initial;
		draw_state->next = NULL;
		// prev should already be NULL

		draw_3D_depth_sort = 0;
	}

	if(draw_sort_mode & SCENESORTZ) {
		d3d_device->SetRenderState(D3DRENDERSTATE_ZENABLE,FALSE); 

		if(d3d_device != NULL) {
			if(draw_sort_mode & SCENESORTS) {
				if(!win_alternator)
					d3d_device->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_GREATEREQUAL);
				else
					d3d_device->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_LESSEQUAL);
			} else {
				d3d_device->Clear(0,NULL,D3DCLEAR_ZBUFFER,0,1.0f,0); // clear the z-buffer
			}
		}
	}

	draw_sort_mode = 0;
}

bool WindowsDX::PrepareDrawState (ImageX * source, bool immediate)
{
	// if drawing normally or storing sorted states
	if(draw_3D_depth_sort == 0 || draw_3D_depth_sort == 1) {
		draw_state->src = source;
		bool ok = GetScreenVertices(draw_state->x,draw_state->y,draw_state->z);
		if(!ok) return false;
		draw_state->zcomp = draw_state->z[0]+draw_state->z[1]+draw_state->z[2]+draw_state->z[3];
	}

	if(draw_3D_depth_sort == 1) { // if depth sorting, store this drawing command (we'll merge-sort the list of stored commands later)

		if(immediate) return true; // unless we're told we don't need to store it
		GameXDrawState * new_state = new GameXDrawState(*draw_state); // make shallow clone of the draw state
		draw_state_num++;
		if(draw_state_head == NULL) {
			// insert at start==end of list
			draw_state_head = new_state;
			draw_state_tail = new_state;
		} else {
			// insert at end of list
			draw_state_tail->next = new_state;
			new_state->prev = draw_state_tail;
			draw_state_tail = new_state;
		}
	}
	return true;
}


// convert draw modes to the closest matching combiner operator
// NOTE: this is COMPLETELY UNRELATED to drawing an image in a certain draw mode like DRAW_ADD,
// this is only used to set how color shading is done, i.e. how the color set by SetDrawShading is used.
// ALSO NOTE: this does not affect how the alpha component of the color is used.
void WindowsDX::SetDrawShadingEffect(DrawFlags effect)
{
	switch(effect & DRAWMODESMASK) {
		case 0: draw_state->color_effect = D3DTOP_SELECTARG1; break; // if no draw mode given, disable shading
		case DRAW_PLAIN: draw_state->color_effect = D3DTOP_SELECTARG2; break; // plain -> the color takes over
		case DRAW_ADD: case DRAW_ADDSHARP: draw_state->color_effect = D3DTOP_ADD; break; // shade color is added to the image being drawn
		case DRAW_GHOST: draw_state->color_effect = D3DTOP_ADDSIGNED; break; // can lighten or darken, usually changes lightness instead of color, often faint
		case DRAW_INTENSIFY: case DRAW_ADDSOFT: draw_state->color_effect = D3DTOP_ADDSMOOTH; break; // similar to with DRAW_ADD
		case DRAW_SUBTRACT: case DRAW_BURN: draw_state->color_effect = D3DTOP_SUBTRACT; break; // shade color subtracted from image -- interesting effect
		case DRAW_INVERT: draw_state->color_effect = D3DTOP_DOTPRODUCT3; break; // not inversion, but it's interesting, albeit black-and-white
		default: case DRAW_MULTIPLY: draw_state->color_effect = D3DTOP_MODULATE; break; // shade color modulates image color -- normally desired effect
	}
}

int WindowsDX::ClipRects (RECT & src_rect, RECT & dst_rect, const RECT & src_bounds, const RECT & dst_bounds)
{
	if(dst_rect.right >= dst_bounds.left && dst_rect.left <= dst_bounds.right) { // In-View Horizontally
		if(dst_rect.bottom >= dst_bounds.top && dst_rect.top <= dst_bounds.bottom) { // In-View Vertically 

			if(dst_rect.right-dst_rect.left == src_rect.right-src_rect.left
			&& dst_rect.bottom-dst_rect.top == src_rect.bottom-src_rect.top) { // If we don't need to scale

				if(dst_rect.left < dst_bounds.left) {
					src_rect.left += dst_bounds.left - dst_rect.left;
					dst_rect.left = dst_bounds.left;
				}
				if(dst_rect.top < dst_bounds.top) {
					src_rect.top += dst_bounds.top - dst_rect.top;
					dst_rect.top = dst_bounds.top;
				}
				if(dst_rect.right > dst_bounds.right) {
					src_rect.right += dst_bounds.right - dst_rect.right;
					dst_rect.right = dst_bounds.right;
				}
				if(dst_rect.bottom > dst_bounds.bottom) {
					src_rect.bottom += dst_bounds.bottom - dst_rect.bottom;
					dst_rect.bottom = dst_bounds.bottom;
				}
				// clip the source too... we could skip this is we were lazy, but then
				// giving an invalid source position would cause the game to crash
				if(src_rect.left < src_bounds.left) {
					dst_rect.left += src_bounds.left - src_rect.left;
					src_rect.left = src_bounds.left;
				}
				if(src_rect.top < src_bounds.top) {
					dst_rect.top += src_bounds.top - src_rect.top;
					src_rect.top = src_bounds.top;
				}
				if(src_rect.right > src_bounds.right) {
					dst_rect.right += src_bounds.right - src_rect.right;
					src_rect.right = src_bounds.right;
				}
				if(src_rect.bottom > src_bounds.bottom) {
					dst_rect.bottom += src_bounds.bottom - src_rect.bottom;
					src_rect.bottom = src_bounds.bottom;
				}
				if(dst_rect.left >= dst_rect.right || dst_rect.top >= dst_rect.bottom) return 0;
				return 1; // Image is in view and should be drawn as-is with the src_rect and dst_rect that we just (possibly) modified
			} else { // need to scale, and if we clip it we need to adjust the clipping amount to match
				
				// the clipping used here isn't perfect since it doesn't allow for pixels to be partially clipped,
				// but it's better than blindly assuming a 1:1 source:destination ratio

				if(dst_rect.left < dst_bounds.left) {
					src_rect.left += (dst_bounds.left - dst_rect.left) * (src_rect.right - src_rect.left) / (dst_rect.right - dst_rect.left);
					dst_rect.left = dst_bounds.left;
				}
				if(dst_rect.top < dst_bounds.top) {
					src_rect.top += (dst_bounds.top - dst_rect.top) * (src_rect.bottom - src_rect.top) / (dst_rect.bottom - dst_rect.top);
					dst_rect.top = dst_bounds.top;
				}
				if(dst_rect.right > dst_bounds.right) {
					src_rect.right += (dst_bounds.right - dst_rect.right) * (src_rect.right - src_rect.left) / (dst_rect.right - dst_rect.left);
					dst_rect.right = dst_bounds.right;
				}
				if(dst_rect.bottom > dst_bounds.bottom) {
					src_rect.bottom += (dst_bounds.bottom - dst_rect.bottom) * (src_rect.bottom - src_rect.top) / (dst_rect.bottom - dst_rect.top);
					dst_rect.bottom = dst_bounds.bottom;
				}
				if(src_rect.left < src_bounds.left) {
					dst_rect.left += (src_bounds.left - src_rect.left) * (dst_rect.right - dst_rect.left) / (src_rect.right - src_rect.left);
					src_rect.left = src_bounds.left;
				}
				if(src_rect.top < src_bounds.top) {
					dst_rect.top += (src_bounds.top - src_rect.top) * (dst_rect.bottom - dst_rect.top) / (src_rect.bottom - src_rect.top);
					src_rect.top = src_bounds.top;
				}
				if(src_rect.right > src_bounds.right) {
					dst_rect.right += (src_bounds.right - src_rect.right) * (dst_rect.right - dst_rect.left) / (src_rect.right - src_rect.left);
					src_rect.right = src_bounds.right;
				}
				if(src_rect.bottom > src_bounds.bottom) {
					dst_rect.bottom += (src_bounds.bottom - src_rect.bottom) * (dst_rect.bottom - dst_rect.top) / (src_rect.bottom - src_rect.top);
					src_rect.bottom = src_bounds.bottom;
				}
				if(dst_rect.left >= dst_rect.right || dst_rect.top >= dst_rect.bottom) return 0;
				return 2; // Image is in view and should be drawn SCALED with the src_rect and dst_rect that we just (possibly) modified
			}
		}
	}
	return 0; // Image is not in view at all and so shouldn't be drawn
}



void WindowsDX::ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy)
{
	if(!xres || !yres)
		return;

	if (x<0) {
		mx = ((int) (x / xres)) - 1;
		offx = (x % xres) + xres;
	} else {
		mx = ((int) (x / xres)) ;
		offx = x % xres;
	}
	if (y<0) {
		my = ((int) (y / yres)) - 1;		
		offy = (y % yres) + yres;
	} else {
		my = ((int) (y / yres)) ;		
		offy = y % yres;
	}	
}

// Params: left, top, right, bottom, red (out of 255), green (out of 255), blue (out of 255)
// (this is the toned-down but ultra-fast version of DrawRect)
void WindowsDX::InternalDrawRectDD (int r, int g, int b, int x1, int y1, int x2, int y2)
{
	if(x2 <= x1) {
		if(draw_state->flags & DRAWOP_NOCULL && x1 != x2) {
			int t = x2;
			x2 = x1;
			x1 = t;
		} else return;
	}
	if(y2 <= y1) {
		if(draw_state->flags & DRAWOP_NOCULL && y1 != y2) {
			int t = y2;
			y2 = y1;
			y1 = t;
		} else return;
	}

	RECT dst_rect = {x1,y1,x2,y2};

	UnlockBackBuffer() ; // unlock drawing surface if necessary

	DDBLTFX ddbltfx;
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = (((r >> (8-GameX.win_maxr)) << win_shiftr) + ((g >> (8-GameX.win_maxg)) << win_shiftg) + ((b >> (8-GameX.win_maxb)) << win_shiftb));

	HRESULT status = dd_back_buf->Blt(&dst_rect, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
	if(FAILED(status))
		DirectDrawError(status);
	else return;
	
	// if we failed, try it the old way:

	LockBackBuffer() ;

	unsigned short *start, *pixel;
	int i, j, clr;

	clr = ddbltfx.dwFillColor;
	start = (unsigned short *) win_screen;
	start += y1*win_pitch + x1;
	for (j=y1; j<=y2; j++) {
		pixel = start;
		for (i=x1; i<=x2; i++)
			*pixel++ = clr;
		start += win_pitch;
	}
}

// 2D drawing functions:
void WindowsDX::DrawPoint	(ColorX & clr, int x, int y)
{
	draw_state->colors = clr; 
	draw_state->vertices_3D = 0;
	draw_state->dst_quad2d = Quad2D((float)x,(float)y,DEFAULT2,DEFAULT2);
	InternalDrawMaster(NULL);
}
void WindowsDX::DrawLine	(ColorX & clr, int x1,int y1, int x2,int y2) 
{
	draw_state->colors = clr;
	draw_state->angle = 0.0f;
	draw_state->vertices_3D = -2;
	draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2);
	InternalDrawMaster(NULL);
}
void WindowsDX::DrawRect	(ColorX & clr, int xLeft, int yTop, int xRight, int yBottom)
{
	draw_state->colors = clr;
	draw_state->vertices_3D = 0;
	draw_state->dst_quad2d = Quad2D((float)xLeft,(float)yTop,(float)xRight,(float)yBottom);
	InternalDrawMaster(NULL);
}

void WindowsDX::DrawPolygon (ColorX & clr, int x1,int y1, int x2,int y2, int x3,int y3, int x4,int y4)
{
	draw_state->colors = clr;
	draw_state->vertices_3D = 0;
	draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2,(float)x3,(float)y3,(float)x4,(float)y4);
	if(!(draw_state->flags & DRAWOP_NOCULL))
		draw_state->flags |= DRAWOP_NOCULL;
	InternalDrawMaster(NULL);
}

// 2D image drawing functions:
void WindowsDX::DrawImage			(ImageX * img, int x, int y)
{
	draw_state->vertices_3D = 0;
	draw_state->dst_quad2d = Quad2D((float)x,(float)y,(float)DEFAULT2,(float)DEFAULT2);
	InternalDrawMaster(img);
}
void WindowsDX::DrawImage			(ImageX * img, int x, int y, float angle, float scale) /* shorthand for calling SetDrawAngle, SetDrawScale before drawing DrawImage */
{
	SetDrawAngle(angle);
	SetDrawScale(scale);
	DrawImage(img,x,y);
}
void WindowsDX::DrawLineImage		(ImageX * img, int x1,int y1, int x2,int y2, bool pixelUnitScale)
{
	draw_state->angle = 0.0f;
	draw_state->vertices_3D = pixelUnitScale ? -2 : -1;
	draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2);
	InternalDrawMaster(img);
}
void WindowsDX::DrawTexturedRect	(ImageX * img, int xLeft, int yTop, int xRight, int yBottom)
{
	draw_state->src = img;
	draw_state->vertices_3D = 0;
	draw_state->dst_quad2d = Quad2D((float)xLeft,(float)yTop,(float)xRight,(float)yBottom);
	InternalDrawMaster(img);
}
void WindowsDX::DrawTexturedPolygon (ImageX * img, int x1,int y1, int x2,int y2,  int x3,int y3,  int x4,int y4)
{
	draw_state->src = img;
	draw_state->vertices_3D = 0;
	draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2,(float)x3,(float)y3,(float)x4,(float)y4);
	if(!(draw_state->flags & DRAWOP_NOCULL)) draw_state->flags |= DRAWOP_NOCULL; // culling of 2D polys would cause confusion
	InternalDrawMaster(img);
}

// 3D drawing functions:
void WindowsDX::DrawPoint3D   (ColorX & clr, Vector3DF & v)
{
	draw_state->colors = clr;
	draw_state->vertices_3D = 3;
	draw_state->dst_point3d = v;
	InternalDrawMaster(NULL);
}
void WindowsDX::DrawLine3D	  (ColorX & clr, Vector3DF & v1, Vector3DF & v2)
{
	draw_state->colors = clr;
	draw_state->vertices_3D = 5;
	draw_state->dst_quad3d = Quad3D(v1,v2,v1,v2);
	InternalDrawMaster(NULL);
}
void WindowsDX::DrawPolygon3D (ColorX & clr, Vector3DF & v1, Vector3DF & v2, Vector3DF & v3, Vector3DF & v4)
{
	draw_state->colors = clr;
	draw_state->vertices_3D = 1;
	draw_state->dst_quad3d = Quad3D(v1,v2,v3,v4);
	InternalDrawMaster(NULL);
}

// 3D image drawing functions:
void WindowsDX::DrawImage3D 		  (ImageX * img, Vector3DF & v, bool worldUnitScale)
{
	draw_state->vertices_3D = worldUnitScale?3:2;
	draw_state->dst_point3d = v;
	InternalDrawMaster(img);
}
void WindowsDX::DrawLineImage3D 	  (ImageX * img, Vector3DF & v1, Vector3DF & v2, bool worldUnitScale)
{
	draw_state->src = img;
	draw_state->vertices_3D = worldUnitScale ? 5 : 4;
	draw_state->dst_quad3d = Quad3D(v1,v2,v1,v2);
	if(draw_state->warp_mode == 0) draw_state->warp_mode = 2; // auto-perspective-correct
	InternalDrawMaster(img);
}
void WindowsDX::DrawTexturedPolygon3D (ImageX * img, Vector3DF & v1, Vector3DF & v2, Vector3DF & v3, Vector3DF & v4)
{
	draw_state->src = img;
	draw_state->vertices_3D = 1;
	draw_state->dst_quad3d = Quad3D(v1,v2,v3,v4);
	if(draw_state->warp_mode == 0) draw_state->warp_mode = 2; // auto-perspective-correct
	InternalDrawMaster(img);
}
void WindowsDX::ClearScreen (void)
{
	if(d3d_main)
		FillScreen(ColorX(0,0,0));
	else
		InternalDrawRectDD(0,0,0, win_viewx1,win_viewy1,win_viewx2,win_viewy2);
}
void WindowsDX::FillScreen	(ColorX & clr)
{
	draw_state->colors = clr;
	draw_state->vertices_3D = 0;
	draw_state->dst_quad2d = Quad2D(0,0,DEFAULT,DEFAULT);
	InternalDrawMaster(NULL);
}
void WindowsDX::BlurScreen	(float amount)
{
	BlurRect(amount, 0,0,DEFAULT,DEFAULT);
}

// fills quadrilateral with screen vertices,
// returns false if it isn't on the screen
bool WindowsDX::GetScreenVertices(float * x, float * y, float * z)
{
	switch(draw_state->vertices_3D) {
		case 0: // regular 2D sprite or rectangle or polygon
			{
				for(int i = 0 ; i < 4 ; i++) {
					x[i] = draw_state->dst_quad2d.x[i];
					y[i] = draw_state->dst_quad2d.y[i];
					z[i] = draw_state->depth2D;
				}

				if(x[3] != DEFAULT2 && (draw_state->scalex != 1.0f || draw_state->scaley != 1.0f)) {
					float xavg = (x[0]+x[1]+x[2]+x[3])/4.0f;
					float yavg = (y[0]+y[1]+y[2]+y[3])/4.0f;
					for(int i = 0 ; i < 4 ; i++) {
						x[i] = ((x[i] - xavg) * draw_state->scalex) + xavg;
						y[i] = ((y[i] - yavg) * draw_state->scaley) + yavg;
					}
				}
			}
			break;
		case 4: // 3D line (line sprite)
		case 5: // 3D line with scale in world units of width instead of as width factor
			{
				Vector3DF top = draw_state->dst_quad3d.v[1];
				Vector3DF bottom = draw_state->dst_quad3d.v[0];
				Vector3DF along = top; along -= bottom;
				
				float relative_width;
				if(draw_state->vertices_3D == 4) {
					relative_width = draw_state->scalex/2.0f;
					if(draw_state->src != NULL) {
						RECT rect = ConvertRECT(draw_state->src_rect, draw_state->src);
						relative_width *= (float)(rect.right-rect.left)/(float)(rect.bottom-rect.top);
					}
				} else {
					float distance = along.Length();
					if(distance == 0.0f)
						return false;
					relative_width = draw_state->scalex / distance;
				}
				if(relative_width == 0.0f) return false;

				relative_width *= along.Length();
				
				Vector3DF into = draw_state->cam->GetDirVector();
				Vector3DF side = Vector3DF(along);
				side.CrossWith(into).Normalize();

				// do special rotation about line axis here
				if(draw_state->angle != 0.0f)
				{
					float angle = draw_state->angle * -DEGtoRAD;
					float sinAng = sinf(angle);
					float cosAng = cosf(angle);

					// rotate the "side" vector about the arbitrary axis "along":
					along.Normalize();
					side = (side*cosAng) + (along*((1.0f-cosAng)*(side.Dot(along)))) + (along.CrossWith(side)*sinAng);

					draw_state->angle = 0.0f; // angle converted to quad coords; prevent extra 2D rotation
				}

				side *= relative_width;

				draw_state->dst_quad3d.v[0] = top+side;
				draw_state->dst_quad3d.v[1] = top-side;
				draw_state->dst_quad3d.v[2] = bottom+side;
				draw_state->dst_quad3d.v[3] = bottom-side;

				draw_state->vertices_3D = 1; // line converted to quad
			}
			// no break, must be followed by case 1
		case 1: // quad of 4 Vector3DF's   (polygon)
			{
				bool ok = true;
				float xd, yd, zd;
				for(int i = 0 ; i < 4 ; i++) {
					
					draw_state->cam->Project(draw_state->dst_quad3d.v[i], xd, yd, zd);

					if(zd < draw_state->cam->GetNear())
						return false;
//					if(zd > draw_state->cam->GetNearClip())
//						ok = true;
//					if(zd < -2000) ok = false;


					x[i] = xd;
					y[i] = yd;
					z[i] = zd;

					if(draw_state->warp_mode == 2) {
						if(zd != 0.0f)	draw_state->warp[i] = 1.0f/(float)zd;
						else			draw_state->warp[i] = 1.0f;
					}
				}
				if(!ok) return false;
				draw_state->angle = 0.0f;
			}
			break;
		case 2: // 3D point, or
		case 3: // 3D point with 3D scaling    (point sprite)
			{
				float xd, yd, zd;			
				draw_state->cam->Project(draw_state->dst_point3d, xd, yd, zd);

				if(zd < draw_state->cam->GetNearClip())
					return false;

				if(draw_state->vertices_3D == 3) {
					float div = 2.0f*max(0.02f,zd)/draw_state->cam->GetPixelsPerUnit();
					float xp = draw_state->scalex/div;
					float yp = draw_state->scaley/div;

					x[0] = xd-xp;
					x[1] = xd+xp;
					x[2] = xd-xp;
					x[3] = xd+xp;
					y[0] = yd-yp;
					y[1] = yd-yp;
					y[2] = yd+yp;
					y[3] = yd+yp;
					z[0] = zd;
					z[1] = zd;
					z[2] = zd;
					z[3] = zd;
				} else {
					x[0] = xd;
					x[2] = xd;
					x[3] = (float)DEFAULT2;
					y[0] = yd;
					y[1] = yd;
					z[0] = zd;
					z[1] = zd;
					z[2] = zd;
					z[3] = zd;
				}

			}
			break;
		case -1: // 2D line (line sprite)
		case -2: // 2D line with scale in pixels of width instead of as width factor
			{
				float xd[2], yd[2];

				xd[0] = draw_state->dst_quad2d.x[0];
				yd[0] = draw_state->dst_quad2d.y[0];
				xd[1] = draw_state->dst_quad2d.x[1];
				yd[1] = draw_state->dst_quad2d.y[2];

				float rise = yd[1]-yd[0];
				float run = xd[0]-xd[1];

				float relative_width;
				if(draw_state->vertices_3D == -1) {
					relative_width = draw_state->scalex/2.0f;
					if(draw_state->src != NULL) {
						RECT rect = ConvertRECT(draw_state->src_rect, draw_state->src);
						relative_width *= (float)(rect.right-rect.left)/(float)(rect.bottom-rect.top);
					}
				} else {
					float distance = sqrtf(run*run+rise*rise);
					if(distance == 0.0f) return false;
					relative_width = draw_state->scalex/2.0f / distance;
				}
				if(relative_width == 0.0f) return false;

				rise *= relative_width;
				run *= relative_width;

				x[0] = xd[1] + rise;
				y[0] = yd[1] + run;
				z[0] = draw_state->depth2D;

				x[1] = xd[1] - rise;
				y[1] = yd[1] - run;
				z[1] = draw_state->depth2D;

				x[2] = xd[0] + rise;
				y[2] = yd[0] + run;
				z[2] = draw_state->depth2D;

				x[3] = xd[0] - rise;
				y[3] = yd[0] - run;
				z[3] = draw_state->depth2D;
			}
			break;
	}

	if(draw_state->dst == VIEWPORT) {
		draw_state->dst_bounds.left = win_viewx1;
		draw_state->dst_bounds.top = win_viewy1;
		draw_state->dst_bounds.right = win_viewx2;
		draw_state->dst_bounds.bottom = win_viewy2;
	} else if(draw_state->dst != NULL) {
		draw_state->dst_bounds.left = 0;
		draw_state->dst_bounds.top = 0;
		draw_state->dst_bounds.right = draw_state->dst->m_xres;
		draw_state->dst_bounds.bottom = draw_state->dst->m_yres;
	}
	if(draw_state->vertices_3D > 0) { // if it's 3D
		int x1, y1, x2, y2;
		draw_state->cam->GetWindow(x1, y1, x2, y2);
		if(x1 > draw_state->dst_bounds.left && x1 < draw_state->dst_bounds.right)
			draw_state->dst_bounds.left = x1;
		if(y1 > draw_state->dst_bounds.top && y1 < draw_state->dst_bounds.bottom)
			draw_state->dst_bounds.top = y1;
		if(x2 > draw_state->dst_bounds.left && x2 < draw_state->dst_bounds.right)
			draw_state->dst_bounds.right = x2;
		if(y2 > draw_state->dst_bounds.top && y2 < draw_state->dst_bounds.bottom)
			draw_state->dst_bounds.bottom = y2;
	}

	if(x[3] == DEFAULT) { // default value for entire destination
		if(draw_state->dst == VIEWPORT) {
			x[0] = x[2] = (float)win_viewx1;
			y[0] = y[1] = (float)win_viewy1;
			x[1] = x[3] = (float)win_viewx2;
			y[2] = y[3] = (float)win_viewy2;
		} else {
			x[1] = (float)draw_state->dst->m_xres;
			y[2] = (float)draw_state->dst->m_yres;
			if(draw_state->dst->m_xmult != 1.0f) x[1] *= draw_state->dst->m_xmult;
			if(draw_state->dst->m_ymult != 1.0f) y[2] *= draw_state->dst->m_ymult;
			x[3] = x[1];
			y[3] = y[2];
		}
	}
	float scalex = draw_state->scalex;
	float scaley = draw_state->scaley;
	if(draw_state->flags & DRAWOP_NOCULL) {
		if(scalex == 0.0f || scaley == 0.0f) return false; // allows 2D negative scale flipping when culling is off
	} else {
		if(scalex <= 0.0f || scaley <= 0.0f) return false;
	}

	RECT sourceRect = ConvertRECT(draw_state->src_rect, draw_state->src);

	if(x[3] == DEFAULT2) { // default value that means we have to convert a point into a rectangle

		float xOff = scalex*(sourceRect.right - sourceRect.left);
		float yOff = scaley*(sourceRect.bottom - sourceRect.top);

		if(draw_state->src != NULL && draw_state->src != VIEWPORT) {
			if(draw_state->src->m_xmult != 1.0f) xOff /= draw_state->src->m_xmult;
			if(draw_state->src->m_ymult != 1.0f) yOff /= draw_state->src->m_ymult;
		}

		x[2] = x[0];
		x[3] = x[0] + xOff;
		x[1] = x[0] + xOff;
		y[3] = y[0] + yOff;
		y[2] = y[0] + yOff;

		if(draw_state->flags & DRAWOP_CENTERED) { // if the point is the center, not the top-left corner, of the destination
			float halfWidth = floorf(((x[1]+x[3]) - (x[0]+x[2])) / 4.0f); // if we make halfWidth a float, it results in higher accuracy but more blurriness with images that should have 1 to 1 pixel to texel ratio
			float halfHeight = floorf(((y[2]+y[3]) - (y[0]+y[1])) / 4.0f);
			for(int i = 0 ; i < 4 ; i++) {
				x[i] = floorf(x[i] - halfWidth);
				y[i] = floorf(y[i] - halfHeight);
			}
			draw_state->flags ^= DRAWOP_CENTERED;
		}
	}

	if(draw_state->transx || draw_state->transy)
	for(int t = 0 ; t < 4 ; t++) {
		x[t] += draw_state->transx;
		y[t] += draw_state->transy;
	}

	float angle = draw_state->angle;
	if(angle != 0.0f) {
		float x_offset = (x[0]+x[1]+x[2]+x[3]) / 4;
		float y_offset = (y[0]+y[1]+y[2]+y[3]) / 4;
		angle = (360.0f-angle)*DEGtoRAD; // convert to CCW radians
		float cosAngle = cosf(angle);
		float sinAngle = sinf(angle);

		// rotate the image by simply rotating the four points
		// of the rectangular polygon that the image is being mapped onto
		// like so:
		//				   >			  0
		//				  / 			/\				
		//	0----1		 0----1 	   /  \ 			
		//	|	 |		 |	  |\	2 /    \	   
		//	|	 |		 |	  | v	 /		\	   
		//	|	 |	   ^ |	  | 	 \		/		 
		//	|	 |		\|	  | 	  \    / 1		   
		//	2----3	->	 2----3   ->   \  / 				
		//					 /			\/				  
		//					<		   3	 

		for(int i = 0 ; i < 4 ; i++) {
			x[i] -= x_offset;
			y[i] -= y_offset;
			float xr = x[i]*cosAngle - y[i]*sinAngle;
			float yr = x[i]*sinAngle + y[i]*cosAngle;
			x[i] = xr + x_offset;
			y[i] = yr + y_offset;
		}
	}

	// catch the things that are completely out of bounds and skip drawing them:
	bool inside = false;
	for(int i = 0 ; i < 4 ; i++) {
		if(x[i] >= draw_state->dst_bounds.left && x[i] <= draw_state->dst_bounds.right && y[i] >= draw_state->dst_bounds.top && y[i] <= draw_state->dst_bounds.bottom) {
			inside = true;
			break;
		}
	}
	if(!inside) {
		// none of the points are in bounds, but that doesn't necessarily mean it's safe to skip drawing it...
		// we'll just do a simple check to see if all of the ones that were out of bounds
		// were on the same side, in which case they couldn't possibly intersect the boundary:
		
		int i;
		bool allout;

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(x[i] >= draw_state->dst_bounds.left) {
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far left; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(y[i] >= draw_state->dst_bounds.top) {
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far up; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(x[i] <= draw_state->dst_bounds.right) {
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far right; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(y[i] <= draw_state->dst_bounds.bottom) {
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far down; don't draw
		
		// if we got here, it *probably* needs to be drawn...
		// here is the only no-draw case we may have missed:
		//
		//			polygon (with image on it)
		//			  /\
		//			 /	\
		//			/	 \
		//		   <	  \
		//		   .\	   >
		//		   . \	  /
		// +--------+ \  /
		// |		|..\/
		// |		|
		// |		|
		// |		|
		// +--------+
		// screen (or other destination boundary)
		// 
		// we could do boundary intersection checks at this point but it wouldn't be worth it.
		// since all completely out of bound images would have to go through the check
		// every single time they're drawn, and the time used for intersection checks could add up quickly
	}

	return true;
}

// convert lowered (from 255) alpha into lowered r,g,b
// used for drawing modes that draw using red, green, blue, effective opacity determined by brightness,
void WindowsDX::ScaleByAlpha (ColorX & colors)
{
	for(int i = 0 ; i < 4 ; i++) {
		if(colors.c[i][3]!=255) {
			for(int j = 0 ; j < 3 ; j++)
				colors.c[i][j] = colors.c[i][j] * colors.c[i][3] / 255;
			colors.c[i][3] = 255;
		}
	}
}

void WindowsDX::SpecialPauseFade(float blur1, float blur2, int r1, int r2, int g1, int g2, int b1, int b2, int a)
{
	bool warn = win_warning_on_draw; SetWarningOnDraw(false);
	int xv1 = win_viewx1, yv1 = win_viewy1, xv2 = win_viewx2, yv2 = win_viewy2;
	SetView(0,0,win_width,win_height);
	BlurScreen(blur1);
	BlurScreen(blur2);
	ColorX c;
	if(win_fullscreen || win_screenbpp != 16)
		c = ColorX(GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a);
	else
		c = ColorX(GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a,
				   GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a,
				   GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a,
				   GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a);
	FillScreen(c);
	SetView(xv1,yv1,xv2,yv2);
	if(warn) SetWarningOnDraw(true);
}

// blurs the target image or the viewport by scaling down then up
// amount of 0.0f means not blurred at all, 0.5f means somewhat blurred, 1.0f means blurred all the way into a single color
void WindowsDX::BlurRect(float amount, int x1, int y1, int x2, int y2)
{
	if(win_warning_on_draw && draw_state->dst == VIEWPORT) {
		ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}

	if(amount <= 0.01f || (!win_blur_supported && draw_state->dst == VIEWPORT)) return;

	if(x2 <= x1 && x2 != DEFAULT) {
		if(draw_state->flags & DRAWOP_NOCULL && x1 != x2) {
			int t = x2;
			x2 = x1;
			x1 = t;
		} else return;
	}
	if(y2 <= y1 && y2 != DEFAULT) {
		if(draw_state->flags & DRAWOP_NOCULL && y1 != y2) {
			int t = y2;
			y2 = y1;
			y1 = t;
		} else return;
	}
	
	amount = (1.0f - amount) * (1.0f - amount); // invert about 0.5 and adjust amount toward 0
	LPDIRECTDRAWSURFACE7 surface;
	if(draw_state->dst == VIEWPORT) {
		surface = dd_back_buf;
		UnlockBackBuffer();
	} else if(draw_state->dst == NULL) {
		return;
	} else {
		surface = draw_state->dst->m_surface;
	}

	if(x2 == DEFAULT) { // blur entire thing
		if(draw_state->dst == VIEWPORT) {
			x1 = win_viewx1;
			y1 = win_viewy1;
			x2 = win_viewx2;
			y2 = win_viewy2;
		} else {
			x2 = draw_state->dst->m_xres;
			y2 = draw_state->dst->m_yres;
			if(draw_state->dst->m_xmult != 1.0f) x2 = (int)((float)x2 * draw_state->dst->m_xmult);
			if(draw_state->dst->m_ymult != 1.0f) y2 = (int)((float)y2 * draw_state->dst->m_ymult);
		}
	}

	int width = (x2-x1);
	int height = (y2-y1);
	RECT full_rect = {x1,y1,x2,y2};
	RECT mini_rect = {x1+(int)(width*(1-amount))/2, y1+(int)(height*(1-amount))/2, x1+(int)(width*(1+amount))/2, y1+(int)(height*(1+amount))/2} ;
	if(mini_rect.right == mini_rect.left) mini_rect.left--;
	if(mini_rect.bottom == mini_rect.top) mini_rect.top--;

	InternalDrawImageDD(surface, surface, full_rect, mini_rect, 1);
	InternalDrawImageDD(surface, surface, mini_rect, full_rect, 1);
}


void WindowsDX::DrawText (int x, int y, char *msg, int r, int g, int b)
{
	if(win_warning_on_draw) {
		ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}

	LockBackBuffer();

	HDC hdc = NULL;
	HRESULT status = dd_back_buf->GetDC(&hdc); // Get the device context handle.
	if(FAILED(status)) {
		DirectDrawError(status);
		return;
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT);
	SetTextColor(hdc, RGB(r,g,b));
	TextOut(hdc, x, y, msg, (int)(strlen(msg))); // just use the Windows function for now, although it's slow

	dd_back_buf->ReleaseDC(hdc);		// Release the device context.
	return;
}

void WindowsDX::DrawOutlinedText (int x, int y, char *msg, int r, int g, int b, int ro, int go, int bo)
{
	if(win_warning_on_draw) {
		ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}

	LockBackBuffer();

	HDC hdc = NULL;
	HRESULT status = dd_back_buf->GetDC(&hdc); // Get the device context handle.
	if(FAILED(status)) {
		DirectDrawError(status);
		return;
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT);

	// Draw an outline around the text
	const int dir[8] = {1,1,1,0,-1,-1,-1,0}; // direction table for outline
	SetTextColor(hdc, RGB(ro,go,bo));
	for (int i=0 ; i<8 ; i++)
		TextOut(hdc, x+dir[i], y+dir[(i+6)%8], msg, (int)(strlen(msg)));

	// Now draw the text normally
	SetTextColor(hdc, RGB(r,g,b));
	TextOut(hdc, x, y, msg, (int)(strlen(msg)));

	dd_back_buf->ReleaseDC(hdc);		// Release the device context.
	return;
}


// draws an image tiled over the entire screen,
// good for things such as drawing backgrounds
void WindowsDX::DrawTiled (ImageX * img, float translation_x, float translation_y, float rotation_center_x, float rotation_center_y)
{
	bool clear_states = !(draw_state->flags & DRAWOP_KEEPSTATES);
	draw_state->flags |= DRAWOP_KEEPSTATES;

	int map_x1, map_y1;
	int map_x2, map_y2; 
	int off_x, off_y;
	float img_x, img_y; 
	float mx, my;

	RECT source_rect = ConvertRECT(draw_state->src_rect, img);

	float xres = draw_state->scalex * (float)(source_rect.right - source_rect.left);
	float yres = draw_state->scaley * (float)(source_rect.bottom - source_rect.top);
	if(img != NULL && draw_state->src != VIEWPORT) {
		if(img->m_xmult != 1.0f) xres /= img->m_xmult;
		if(img->m_ymult != 1.0f) yres /= img->m_ymult;
	}
	if(xres < 2 || yres < 2) return;

	ClipTile ((int)translation_x+win_viewx2,(int)translation_y+win_viewy2, (int)xres, (int)yres, map_x2, map_y2, off_x, off_y) ;
	ClipTile ((int)translation_x+win_viewx1,(int)translation_y+win_viewy1, (int)xres, (int)yres, map_x1, map_y1, off_x, off_y) ;

	float map_numx = (float)(map_x2 - map_x1 + 1);
	float map_numy = (float)(map_y2 - map_y1 + 1);

	if(draw_state->angle == 0.0f) {
		draw_state->vertices_3D = 0;
		for (my = 0; my < map_numy; my++) {
			img_y = my * yres - off_y;
			img_x = (float)(-off_x);
			for (mx = 0; mx < map_numx; mx++) {
				draw_state->dst_quad2d = Quad2D(img_x,img_y,DEFAULT2,DEFAULT2);
				InternalDrawMaster(img);
				img_x += xres;
			}
		}
	} else {

		if(rotation_center_x == DEFAULT)
			rotation_center_x = (float)(win_viewx1+win_viewx2)/2;
		if(rotation_center_y == DEFAULT)
			rotation_center_y = (float)(win_viewy1+win_viewy2)/2;

		float sinAngle, cosAngle;

		sinAngle = sinf((360-draw_state->angle)*DEGtoRAD);
		cosAngle = cosf((360-draw_state->angle)*DEGtoRAD);

		map_numx *= 3;
		map_numy *= 3;

		draw_state->vertices_3D = 0;
		for (my = -map_numy; my < map_numy; my++) {
			img_y = my * yres;
			img_x = -map_numx * xres;
			for (mx = -map_numx; mx < map_numx; mx++) {
				float real_x, real_y, x_offset, y_offset;
				x_offset = rotation_center_x - xres/2;
				y_offset = rotation_center_y - yres/2;
				img_x -= x_offset;
				img_y -= y_offset;
				real_x = img_x * cosAngle - img_y * sinAngle - translation_x;
				real_y = img_x * sinAngle + img_y * cosAngle - translation_y;
				real_x += x_offset;
				real_y += y_offset;
				img_x += x_offset;
				img_y += y_offset;
				draw_state->dst_quad2d = Quad2D(real_x,real_y,DEFAULT2,DEFAULT2);
				InternalDrawMaster(img);
				img_x += xres;
			}
		}
	}
	if(clear_states)
		ResetDrawStates();

	return;
}

//*** Switched Drawing Routines
static void DrawPixel16 (int x, int y, int r, int g, int b, int a=255); // Sets pixel (16-bit) -- r,g,b,a out of 255
static void DrawPixel32 (int x, int y, int r, int g, int b, int a=255); // Sets pixel (32-bit) -- r,g,b,a out of 255
static void ReadPixel16 (int x, int y, int &r, int &g, int &b, int &a); // Gets pixel (16-bit) -- r,g,b,a out of 255
static void ReadPixel32 (int x, int y, int &r, int &g, int &b, int &a); // Gets pixel (32-bit) -- r,g,b,a out of 255
// For images that must be scaled due to limitations:
static void DrawScaledPixel16 (int x, int y, int r, int g, int b, int a=255);
static void DrawScaledPixel32 (int x, int y, int r, int g, int b, int a=255);
static void ReadScaledPixel16 (int x, int y, int &r, int &g, int &b, int &a);
static void ReadScaledPixel32 (int x, int y, int &r, int &g, int &b, int &a);
// So we can error-check the usage of drawpixel and readpixel:
static void DrawPixelError (int x, int y, int r, int g, int b, int a=255);
static void ReadPixelError (int x, int y, int &r, int &g, int &b, int &a);


void WindowsDX::AccessPixels (ImageX * destination)
{
	if(win_warning_on_draw && destination == VIEWPORT) {
		ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}
	if(destination == VIEWPORT || destination == NULL) {
		win_pixel_dest_surface = dd_back_buf;
		win_pixel_dest_data = win_screen;
		win_pixel_dest_pitch = win_adjustedpitch;
		win_pixel_dest_ytable = win_ytable;
		LockBackBuffer();

		if(win_screenbpp == 32) {
			FuncDrawPixel = &DrawPixel32;
			FuncReadPixel = &ReadPixel32;
		} else {
			FuncDrawPixel = &DrawPixel16;
			FuncReadPixel = &ReadPixel16;
		}
	} else {
		win_pixel_dest_surface = destination->m_surface;
		
		if(win_pixel_dest_surface == NULL) {
			win_pixel_dest_data = (char *) destination->m_data;
			win_pixel_dest_pitch = destination->m_xres*8/win_screenbpp;
			win_pixel_dest_ytable = destination->m_ytable;

			if(win_screenbpp == 32) {
				FuncDrawPixel = &DrawPixel32;
				FuncReadPixel = &ReadPixel32;
			} else {
				FuncDrawPixel = &DrawPixel16;
				FuncReadPixel = &ReadPixel16;
			}
		}
		else {
			if(destination->m_surf_bpp == 32) {
				if(destination->m_xmult == 1.0f && destination->m_ymult == 1.0f) {
					FuncDrawPixel = &DrawPixel32;
					FuncReadPixel = &ReadPixel32;
				} else {
					FuncDrawPixel = &DrawScaledPixel32;
					FuncReadPixel = &ReadScaledPixel32;
					win_pixel_dest_xmult = destination->m_xmult;
					win_pixel_dest_ymult = destination->m_ymult;
				}
			} else {
				if(destination->m_xmult == 1.0f && destination->m_ymult == 1.0f) {
					FuncDrawPixel = &DrawPixel16;
					FuncReadPixel = &ReadPixel16;
				} else {
					FuncDrawPixel = &DrawScaledPixel16;
					FuncReadPixel = &ReadScaledPixel16;
					win_pixel_dest_xmult = destination->m_xmult;
					win_pixel_dest_ymult = destination->m_ymult;
				}
			}

			DDSURFACEDESC2 ddsd;
			ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2)) ;
			ddsd.dwSize = sizeof(ddsd) ;
			DirectDrawError(win_pixel_dest_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL));
			win_pixel_dest_data = (char *) ddsd.lpSurface;
			win_pixel_dest_pitch = ddsd.lPitch * 8 / destination->m_surf_bpp;
			win_pixel_dest_ytable = destination->m_ytable;

			destination->m_non_empty = true;
		}
	}
}

void WindowsDX::EndPixelAccess (void)
{
	if(win_pixel_dest_data != win_screen && win_pixel_dest_surface != NULL) {
		win_pixel_dest_surface->Unlock(NULL);
	}
	FuncDrawPixel = &DrawPixelError;
	FuncReadPixel = &ReadPixelError;
}

// this function is very slow if compiler optimizations are off (i.e. in debug mode)
// because it is made to be inlined.
static void DrawPixel16 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
//	((unsigned short *)GameX.win_pixel_dest_data)[x+GameX.win_ytable[y]] = (((r >> (8-GameX.win_maxr)) << GameX.win_shiftr) + ((g >> (8-GameX.win_maxg)) << GameX.win_shiftg) + ((b >> (8-GameX.win_maxb)) << GameX.win_shiftb));
	((unsigned short *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]] = (((r >> 3) << GameX.win_shiftr) + ((g >> (8-GameX.win_maxg)) << 5) + (b >> 3)); // optimized
}

static void DrawPixel32 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
//	((unsigned long *)GameX.win_pixel_dest_data)[x+GameX.win_ytable[y]] = (((r >> (8-GameX.win_maxr)) << GameX.win_shiftr) + ((g >> (8-GameX.win_maxg)) << GameX.win_shiftg) + ((b >> (8-GameX.win_maxb)) << GameX.win_shiftb));
	((unsigned long *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]] = ((a << 24) + (r << 16) + (g << 8) + b); // optimized
}

static void ReadPixel16 (int x, int y, int &r, int &g, int &b, int &a)
{
	unsigned short v = ((unsigned short *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]];
	b =  (v & 0x1F) << 3;
	g = ((v & GameX.win_maskg) >> 5) << (8-GameX.win_maxg) ;
	r = ((v & GameX.win_maskr) >> GameX.win_shiftr) << 3;
	a = 255;
}

static void ReadPixel32 (int x, int y, int &r, int &g, int &b, int &a)
{
	unsigned long v = ((unsigned long *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]];
	b =  v & 0x000000FF;
	g = (v & 0x0000FF00) >> 8;
	r = (v & 0x00FF0000) >> 16;
	a = (v & 0xFF000000) >> 24;
}

static void DrawPixelError (int x, int y, int r, int g, int b, int a)
{
	GameX.ReportProblem(GAMEX_PIXEL_ACCESS_FAILURE);
}

static void ReadPixelError (int x, int y, int &r, int &g, int &b, int &a)
{
	GameX.ReportProblem(GAMEX_PIXEL_ACCESS_FAILURE);
}

// pixel setter/getter function to be used when the destination is smaller than it's pretending to be

static void DrawScaledPixel16 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
	// prevent multiple writes to the same pixel:
	if(GameX.win_pixel_dest_xmult != 1.0f && x != 0 && (int)(GameX.win_pixel_dest_xmult * (float)(x-1)) == (int)(GameX.win_pixel_dest_xmult * (float)(x))) return;
	if(GameX.win_pixel_dest_ymult != 1.0f && y != 0 && (int)(GameX.win_pixel_dest_ymult * (float)(y-1)) == (int)(GameX.win_pixel_dest_ymult * (float)(y))) return;

	// scale the x axis (the y axis is already scaled in the ytable):
	x = (int)((float)x * GameX.win_pixel_dest_xmult);

	// set the pixel
	((unsigned short *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]] = (((r >> 3) << GameX.win_shiftr) + ((g >> (8-GameX.win_maxg)) << 5) + (b >> 3)); // optimized
}

static void DrawScaledPixel32 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
	// prevent multiple writes to the same pixel:
	if(GameX.win_pixel_dest_xmult != 1.0f && x != 0 && (int)(GameX.win_pixel_dest_xmult * (float)(x-1)) == (int)(GameX.win_pixel_dest_xmult * (float)(x))) return;
	if(GameX.win_pixel_dest_ymult != 1.0f && y != 0 && (int)(GameX.win_pixel_dest_ymult * (float)(y-1)) == (int)(GameX.win_pixel_dest_ymult * (float)(y))) return;

	// scale the x axis (the y axis is already scaled in the ytable):
	x = (int)((float)x * GameX.win_pixel_dest_xmult);

	// set the pixel
	((unsigned long *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]] = ((a << 24) + (r << 16) + (g << 8) + b); // optimized
}

static void ReadScaledPixel16 (int x, int y, int &r, int &g, int &b, int &a)
{
	x = (int)((float)x * GameX.win_pixel_dest_xmult);
	unsigned short v = ((unsigned short *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]];
	b =  (v & 0x1F) << 3;
	g = ((v & GameX.win_maskg) >> 5) << (8-GameX.win_maxg) ;
	r = ((v & GameX.win_maskr) >> GameX.win_shiftr) << 3;
	a = 255;
}

static void ReadScaledPixel32 (int x, int y, int &r, int &g, int &b, int &a)
{
	x = (int)((float)x * GameX.win_pixel_dest_xmult);
	unsigned long v = ((unsigned long *)GameX.win_pixel_dest_data)[x+GameX.win_pixel_dest_ytable[y]];
	b =  v & 0x000000FF;
	g = (v & 0x0000FF00) >> 8;
	r = (v & 0x00FF0000) >> 16;
	a = (v & 0xFF000000) >> 24;
}


void WindowsDX::InternalDrawMaster(ImageX * source)
{
	if(win_err_debug) {
		if(win_warning_on_draw && draw_state->dst == VIEWPORT) {
			ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
			ResetDrawStates();
			return;
		}
	}

	if(!(draw_state->flags & DRAWMODESMASK))
		draw_state->flags |= DRAW_PLAIN; // assume no drawing mode means plain drawing mode

	if(draw_state->dst != VIEWPORT && !(draw_state->dst->m_usage & LOAD_TARGETABLE))
		draw_state->dst->ConvertUsageTo(draw_state->dst->m_usage | LOAD_TARGETABLE); // auto-convert to targetable if needed

	bool skip = false;
	bool immediate = false;

	if(draw_3D_depth_sort == 2) { // replaying drawing
		source = draw_state->src;
	} else {
		if(draw_sort_mode & SCENESORTZ && draw_sort_mode & SCENESORTD) {
			if(draw_state->flags & DRAW_PLAIN && !draw_state->colors.HasTransparency() && (source == NULL || !(source->m_usage & LOAD_ALPHA)))
				immediate = true;
		}
		skip = !PrepareDrawState(source,immediate);
	}
	if(draw_3D_depth_sort == 1 && !immediate)
		skip = true;

	if(source != NULL && source != VIEWPORT && source->status != IMG_OK) {
		ReportProblem(GAMEX_UNINIT_IMG_DRAW_FAIL);
		skip = true;
	}

	RECT sourceRect;
	float * x, * y, * z;
	if(!skip) {
		if(draw_state->src != NULL)
			sourceRect = ConvertRECT(draw_state->src_rect, draw_state->src);

		x = draw_state->x; // pointer to an array of 4 variable-purpose x screen coordinates
		y = draw_state->y; // pointer to an array of 4 variable-purpose y screen coordinates
		z = draw_state->z; // used for depth sorting

		draw_state->colors.CapComponents();

		if(!(draw_state->flags & DRAWOP_NOBLENDALPHA)) {
			if(!(draw_state->flags & DRAW_PLAIN)) {
				// scale the luminosity by the alpha which would otherwise be ignored
				if(draw_state->dst == VIEWPORT || !(draw_state->dst->m_usage & LOAD_ALPHA)) {
					ScaleByAlpha(draw_state->colors);
					if(draw_state->color_effect == D3DTOP_MODULATE)
						if(!(draw_state->colors.HasColor()))
							skip = true; // return if there isn't any luminosity
				}
			} else {
				if(!(draw_state->colors.HasOpacity()))
					skip = true; // return if there isn't any opacity
			}
		}
	}

	if(skip) {
		if(!(draw_state->flags & DRAWOP_KEEPSTATES))
			ResetDrawStates();
		return;
	}

	int sourceHasSurface, destHasSurface;
	bool sourceHasData, sourceHasAlpha, destHasData, destHasAlpha, destHasDevice;
	LPDIRECTDRAWSURFACE7 p_src_surface, p_dst_surface;
	LPDIRECT3DDEVICE7 p_dst_device;
	XBYTE * p_src_data;
	XBYTE * p_src_alpha;
	XBYTE * p_dst_data;
	XBYTE * p_dst_alpha;
	RECT src_bounds = {0,0,0,0};
	int src_pitch, dst_pitch;

	if(draw_state->src == VIEWPORT) {
		sourceHasSurface = 1;
		p_src_surface = dd_back_buf;
		sourceHasAlpha = false;
		src_pitch = win_adjustedpitch;
		src_bounds.right = win_width;
		src_bounds.bottom = win_height;
	} else if(draw_state->src != NULL) {
		if(draw_state->src->status != IMG_OK) return;
		if(draw_state->src->m_hassurface) {
			sourceHasSurface = draw_state->src->m_hassurface;
			p_src_surface = draw_state->src->m_surface;
		}
		else sourceHasSurface = 0;

		if(draw_state->src->m_data != NULL) {
			sourceHasData = true;
			p_src_data = draw_state->src->m_data;
		}
		else
			sourceHasData = false;

		if(draw_state->src->m_alpha != NULL) {
			sourceHasAlpha = true;
			p_src_alpha = draw_state->src->m_alpha;
		}
		else
			sourceHasAlpha = false;

		src_pitch = draw_state->src->m_xres; // may have to change this line to do depths other than 16 bit
		src_bounds.right = draw_state->src->m_xres;
		src_bounds.bottom = draw_state->src->m_yres;
		if(draw_state->src->m_xmult != 1.0f) src_bounds.right = (LONG)((float)src_bounds.right * draw_state->src->m_xmult);
		if(draw_state->src->m_ymult != 1.0f) src_bounds.bottom = (LONG)((float)src_bounds.bottom * draw_state->src->m_ymult);
	} else {
		sourceHasSurface = 2;
		p_src_surface = NULL;
		sourceHasData = false;
		sourceHasAlpha = false;
	}

	if(draw_state->dst == VIEWPORT) {
		destHasSurface = 3;
		p_dst_surface = dd_back_buf;

		destHasDevice = true;
		p_dst_device = d3d_device;

		destHasAlpha = false;

		dst_pitch = win_adjustedpitch;
	} else {
		if(draw_state->dst->status != IMG_OK) return;
		if(draw_state->dst->m_hassurface) {
			destHasSurface = draw_state->dst->m_hassurface;
			p_dst_surface = draw_state->dst->m_surface;
		}
		else {
			dst_pitch = draw_state->dst->m_xres; // may have to change this line to do depths other than 16 bit...maybe
			destHasSurface = 0;
		}

		if(draw_state->dst->m_data != NULL) {
			destHasData = true;
			p_dst_data = draw_state->dst->m_data;
		}
		else
			destHasData = false;

		if(draw_state->dst->m_alpha != NULL) {
			destHasAlpha = true;
			p_dst_alpha = draw_state->dst->m_alpha;
		}
		else
			destHasAlpha = false;

		if(draw_state->dst->m_device != NULL) {
			destHasDevice = true;
			p_dst_device = draw_state->dst->m_device;
		}
		else
			destHasDevice = false;
	}

	int alphatype = 0;
	if(draw_state->src != VIEWPORT && draw_state->src != NULL) {
		if((draw_state->src->m_usage & LOAD_MASKED))
			alphatype = 1;
		else if((draw_state->src->m_usage & LOAD_ALPHA)) {
			if(draw_state->dst == VIEWPORT || !(draw_state->dst->m_usage & LOAD_ALPHA))
			alphatype = 2; // alpha channel in image
		}
	}

	// NOTE: support for all advanced features is not complete for if we can't draw using 3D,
	//		 but that *should* only happen when drawing FROM the back buffer to somewhere,

	int ok = 0;
	if(sourceHasSurface==2 && destHasDevice) {
		if(p_src_surface)
			ok = InternalDrawImageD3D(p_dst_device, p_src_surface, sourceRect, src_bounds, x, y, z, draw_state->src->m_surf_width, draw_state->src->m_surf_height, alphatype);
		else
			ok = InternalDrawImageD3D(p_dst_device, NULL, FULLRECT, FULLRECT, x, y, z, win_width, win_height, alphatype);
	}
	if(!ok) {
///		bool can_2D_accelerate = (sourceHasSurface && destHasSurface && p_src_surface != NULL && (draw_state->flags & DRAW_PLAIN) && alphatype != 2 && draw_state->angle==0.0f && !draw_state->colors.HasShading() && (draw_state->vertices_3D==0||draw_state->vertices_3D==2||draw_state->vertices_3D==3));
		bool can_2D_accelerate = destHasSurface ? true : false;

		RECT destRect;
		destRect.left = (LONG)(x[0]+x[2])/2;
		destRect.top = (LONG)(y[0]+y[1])/2;
		destRect.right = (LONG)(x[1]+x[3])/2;
		destRect.bottom = (LONG)(y[2]+y[3])/2;

		if(can_2D_accelerate) {
			if(draw_state->src == VIEWPORT || draw_state->dst == VIEWPORT)
				UnlockBackBuffer();
			if(dd_was_lost) {
				if(draw_state->src && draw_state->src != VIEWPORT) {p_src_surface = draw_state->src->m_surface;}
				if(draw_state->dst && draw_state->dst != VIEWPORT) {p_dst_surface = draw_state->dst->m_surface;}
			}
			if(sourceHasSurface && p_src_surface != NULL) // plain directdraw draw
				ok = InternalDrawImageDD(p_src_surface, p_dst_surface, sourceRect, destRect, !alphatype);
			else { // solid fill color
				InternalDrawRectDD(draw_state->colors.c[0][0],draw_state->colors.c[0][1],draw_state->colors.c[0][2],
									destRect.left, destRect.top, destRect.right, destRect.bottom);
				ok = true;
			}
		}
		if(!ok) {
			// insert software rendering backup code here
		}
	}

	if(draw_state->dst != VIEWPORT && destHasSurface == 2) {
		// if the destination is a texture, we have to tell D3D to refresh it by
		// locking and unlocking it, or else the image won't change after the first time it's drawn.
		// note that this makes drawing to somewhere else than the screen even slower...
		DDSURFACEDESC2 ddsd;
		ddsd.dwSize = sizeof(ddsd);
		p_dst_surface->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
		p_dst_surface->Unlock(NULL);
	}
	if(draw_3D_depth_sort != 2)
		if(!(draw_state->flags & DRAWOP_KEEPSTATES))
			ResetDrawStates();

	if(win_request_slow_debug_drawing && win_init_done && draw_state->dst == VIEWPORT) {
		if((GetAsyncKeyState(VK_ESCAPE) & 0x8000)) {
			win_request_slow_debug_drawing = false; // must be false before choice dialog
			if(ChoiceDialog("Stop drawing slowly?") != 1)
				win_request_slow_debug_drawing = true;
		} else {
			UpdateView();
			Sleep(60);
		}
	}
} // end of InternalDrawMaster

static D3DVIEWPORT7 view;
int lastSrcBlend, lastDestBlend;
int lastShadeCombineEffect;

// Draws a 2D image using 3D acceleration and features
int WindowsDX::InternalDrawImageD3D(LPDIRECT3DDEVICE7 device, LPDIRECTDRAWSURFACE7 image, RECT & src_rect, RECT & src_bounds, float * x, float * y, float * z, int surf_width, int surf_height, int alphatype)
{
	if(device == NULL) return 0;

	if(draw_state->dst == VIEWPORT) {
		UnlockBackBuffer(); // must be "closed" (actually called unlocked) to draw with D3D (or DD for that matter)
	}

	if(d3d_device != NULL
		&& (draw_state->dst_bounds.left != view.dwX // it turns out the SetViewport() is really quite slow,
		 || draw_state->dst_bounds.top != view.dwY  // so these extra checks can greatly improve performance
		 || draw_state->dst_bounds.right-draw_state->dst_bounds.left != view.dwWidth // when optimizations are on
		 || draw_state->dst_bounds.bottom-draw_state->dst_bounds.top != view.dwHeight)) // especially when drawing lots of things
	{
		view.dwX = draw_state->dst_bounds.left;
		view.dwY = draw_state->dst_bounds.top;
		view.dwWidth = draw_state->dst_bounds.right-draw_state->dst_bounds.left;
		view.dwHeight = draw_state->dst_bounds.bottom-draw_state->dst_bounds.top; 
		view.dvMinZ = 0.0f;
		view.dvMaxZ = 1.0f;
		d3d_device->SetViewport(&view);
	}
	// Because Direct3D does not support quadrilateral primitives,
	// create vertices for our quad with points ordered like so:
	//
	//	0----1
	//	.	/.
	//	.  / .
	//	. /  .
	//	./	 .
	//	2----3
	//
	// to allow for D3D to use it for a triangle strip.
	float angle = draw_state->angle;
	int angle_adjust_counter = 999; // prevent infinite angle adjusting when divide-by-zero is passed as angle
	while(angle >= 360.0f && angle_adjust_counter) angle -= 360.0f, angle_adjust_counter--; // restrict to 0-360 range to allow easier angle checking
	while(angle < 0.0f	  && angle_adjust_counter) angle += 360.0f, angle_adjust_counter--; // and also increase cosine and sine function accuracy
	int rotate;
	if(angle==	0.0f || angle== 45.0f || angle== 90.0f || angle==135.0f
	|| angle==180.0f || angle==225.0f || angle==270.0f || angle==315.0f)
		rotate = 0; // if the angle is exactly a multiple of 45, we don't have to adjust its inset as much
	else
		rotate = 1;

	float x_adjust, y_adjust; // source inset amount
	float leftMap, topMap, rightMap, bottomMap;

	if(&src_bounds != &FULLRECT) {
		if(src_rect.right - src_rect.left == src_bounds.right - src_bounds.left) {
			x_adjust = 0.0005f;
		} else {
			float scalex = (float)(x[3]-x[0])/(src_rect.right-src_rect.left);
			x_adjust = ((scalex <= 1.0f) ? (0.0005f) : (0.525f-0.5f/scalex));
			if(rotate) x_adjust = min(0.5f, x_adjust+0.25f);
		}
		if(src_rect.bottom - src_rect.top == src_bounds.bottom - src_bounds.top) {
			y_adjust = 0.0005f;
		} else {
			float scaley = (float)(y[3]-y[0])/(src_rect.bottom-src_rect.top);
			y_adjust = ((scaley <= 1.0f) ? (0.0005f) : (0.525f-0.5f/scaley));
			if(rotate) y_adjust = min(0.5f, y_adjust+0.25f);
		}

		// respond to mirroring request flags
		if(draw_state->flags & DRAWOP_VMIRROR) { // flip the top and bottom mapping values if we want to mirror the image vertically
			int temp = src_rect.top;
			src_rect.top = src_rect.bottom;
			src_rect.bottom = temp;
			y_adjust = -y_adjust;
		}
		if(draw_state->flags & DRAWOP_HMIRROR) { // flip the left and right mapping values if we want to mirror the image horizontally
			int temp = src_rect.left;
			src_rect.left = src_rect.right;
			src_rect.right = temp;
			x_adjust = -x_adjust;
		}

		leftMap   = (src_rect.left	 + x_adjust) / surf_width;	// determine source mapping values
		topMap	  = (src_rect.top	 + y_adjust) / surf_height; // (the pixel adjustments fix a
		rightMap  = (src_rect.right  - x_adjust) / surf_width;	//	"sometimes off by one" bug)
		bottomMap = (src_rect.bottom - y_adjust) / surf_height;
	} else {
		leftMap  = 0.0f; topMap    = 0.0f;
		rightMap = 1.0f; bottomMap = 1.0f;
	}

	int color_shift = 0; // amount to left-shift brightness

	if(draw_state->flags & DRAWOP_BRIGHT)
		color_shift++; // multiply by 2 if BRIGHT

	// Determine which workarounds, if any, we need to apply right now:
	bool work_inten = win_intensify_workaround	 == 1 && draw_state->flags & DRAW_INTENSIFY;
	bool work_subt1 = win_subtract_workaround	 == 1 && draw_state->flags & DRAW_SUBTRACT && alphatype != 2;
	bool work_subt2 = win_subtract_workaround	 == 2 && draw_state->flags & DRAW_SUBTRACT;
	bool work_alpha = win_alpha_blend_workaround == 1 && draw_state->flags & DRAW_PLAIN && alphatype != 2 && draw_state->colors.HasTransparency();
	bool work_ghost = win_ghost_workaround		 == 1 && draw_state->flags & DRAW_GHOST;
	bool work_burn	= win_burn_workaround		 == 1 && draw_state->flags & DRAW_BURN;
	bool work_addso = win_addsoft_workaround	 == 1 && draw_state->flags & DRAW_ADDSOFT;
	bool work_addsh = win_addsharp_workaround	 == 1 && draw_state->flags & DRAW_ADDSHARP;

	// Implement work-around methods as necessary:
	// (these are to work around problems with older video cards, as decided in DoInitTests())
	if(work_inten) {
		draw_state->flags ^= DRAW_ADD | DRAW_INTENSIFY; // switch from intensify to regular additive mode
		draw_state->colors /= 2.0f; // cut brightness in half
	}
	if(work_addso)
		draw_state->flags ^= DRAW_ADD | DRAW_ADDSOFT; // switch from soft add to regular additive mode
	if(work_addsh)
		draw_state->flags ^= DRAW_ADD | DRAW_ADDSHARP; // switch from sharp add to regular additive mode
	if(work_burn) // note that part of the burn workaround happens later, also
		draw_state->flags ^= DRAW_INVERT | DRAW_BURN; // switch from burn to inversion
	if(work_subt1 || work_subt2 || work_ghost || work_alpha) {
		if(image == NULL || image == dd_back_buf) {
			if(work_subt1 || work_subt2) {
				image = NULL;
				draw_state->flags ^= DRAW_PLAIN | DRAW_SUBTRACT; // switch from subtractive to regular alpha mode
				for(int i = 0 ; i < 4 ; i++) {
					draw_state->colors.c[i][3] = (int)((float)draw_state->colors.c[i][0]*RED_LUMINANCE + (float)draw_state->colors.c[i][1]*GREEN_LUMINANCE + (float)draw_state->colors.c[i][2]*BLUE_LUMINANCE) * draw_state->colors.c[i][3]/255;
					draw_state->colors.c[i][2] = 0;
					draw_state->colors.c[i][1] = 0;
					draw_state->colors.c[i][0] = 0;
				}
			}
		} else {
			int compat_index = (work_ghost) ? 2 : ((work_subt2) ? 1 : 0);
			ImageX * work_img = draw_state->src->m_internal_compatibility_img[compat_index];
			if(work_img == NULL) {
				SetStatus(GAMEX_CREATING_TEST_SURFACE); // prevent from registering as user-created
				work_img = new ImageX();
				draw_state->src->m_internal_compatibility_img[compat_index] = work_img;
				work_img->Create(draw_state->src->m_xres, draw_state->src->m_yres, true);
				SetStatus(GAMEX_READY);
				draw_state->src->CopyTo(work_img);
				if(work_subt2 || work_ghost) {
					AccessPixels(work_img);
					for(int y = 0 ; y < work_img->m_yres ; y++) {
						for(int x = 0 ; x < work_img->m_xres ; x++) {
							int r,g,b,a;
							ReadPixel(x,y,r,g,b,a);
							a = (int)((float)r*RED_LUMINANCE + (float)g*GREEN_LUMINANCE + (float)b*BLUE_LUMINANCE) * a/255;
							if(work_subt2) r = 0, g = 0, b = 0;
							DrawPixel(x,y,r,g,b,a);
						}
					}
					EndPixelAccess();
				}
			}
			image = work_img->m_surface;
			if(work_subt2 || work_ghost) {
				draw_state->flags |= DRAWMODESMASK, draw_state->flags ^= DRAWMODESMASK; // turn off all draw modes
				alphatype = 2, draw_state->flags | DRAW_PLAIN; // turn on alpha with plain drawing mode
				for(int i = 0 ; i < 4 ; i++) // scale shading transparency with brightness
					draw_state->colors.c[i][3] = min(255, (int)((float)draw_state->colors.c[i][0]*RED_LUMINANCE + (float)draw_state->colors.c[i][1]*GREEN_LUMINANCE + (float)draw_state->colors.c[i][2]*BLUE_LUMINANCE) *2* draw_state->colors.c[i][3]/255);
			}
		}
	}

	int srcBlend, destBlend;

	switch(draw_state->flags & DRAWMODESMASK) {
		default:
		case DRAW_PLAIN: // plain, masked, or alpha drawing
			if(alphatype != 2 && !draw_state->colors.HasTransparency()) {
				srcBlend = D3DBLEND_ONE; // Cs = 1
				destBlend = D3DBLEND_ZERO; // Cd = 0
				// so C = source * Cs + dest * Cd
				//		= source * (1) + dest * (0)
				//		= source
			} else {
				srcBlend = D3DBLEND_SRCALPHA; // Cs = alpha
				destBlend = D3DBLEND_INVSRCALPHA; // Cd = 1-alpha
				// so C = source * Cs + dest * Cd
				//		= source * (alpha) + dest * (1-alpha)
				//		= dest + source*alpha - dest*alpha
			}
			break;
		case DRAW_ADD: // additive -- r, g, b at 255 is full additive, reduce r for cyan-tinted added, etc.
			if(alphatype != 2) srcBlend = D3DBLEND_ONE; // Cs = 1, or
			else			   srcBlend = D3DBLEND_SRCALPHA; // Cs = alpha
			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (1 [* alpha]) + dest * (1)
			//		= dest + source [* alpha]
			break;
		case DRAW_SUBTRACT: // subtractive -- r, g, b at 255 is full subtractive, reduce r to subtract less red, etc.
			// note: not totally accurate subtraction because this does not use DirectX 8.0
			//		 so the closest thing in DX7 is this, the inverse of DRAW_INTENSIFY
			srcBlend = D3DBLEND_ZERO; // Cs = 0, or
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (0) + dest * (1-source)
			//		= dest - source*dest
			color_shift++; // increase effect 2x because it's usually too pale otherwise (so we don't have to render it twice)
			//	  C = dest - (2*source)*dest
			break;
		case DRAW_GHOST: // color varying alpha blend -- r, g, b determine amount, produces interesting auto-alpha effect
			srcBlend = D3DBLEND_SRCCOLOR; // Cs = source
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (source) + dest * (1-source)
			//		= dest + source*source - source*dest
			break;
		case DRAW_INVERT: // invert by the amount of color in the source -- r,g,b at 255 is full inversion, otherwise hard to describe
			srcBlend = D3DBLEND_INVDESTCOLOR; // Cs = 1-dest
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (1-dest) + dest * (1-source)
			//		= dest + 2*source - 2*source*dest
			break;
		case DRAW_MULTIPLY: // multiply source by destination
			srcBlend = D3DBLEND_ZERO; // Cs = 0
			destBlend = D3DBLEND_SRCCOLOR; // Cd = source
			// so C = source * Cs + dest * Cd
			//		= source * (0) + dest * (source)
			//		= dest * source
			break;
		case DRAW_BURN: // similar to invert but peaks at midway brightness
			srcBlend = D3DBLEND_INVSRCCOLOR; // Cs = 1-source
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (1-source) + dest * (1-source)
			//		= source - source*source + dest - source*dest
			break;
		case DRAW_INTENSIFY: // add, but darker as destination becomes darker
			srcBlend = D3DBLEND_DESTCOLOR; // Cs = dest
			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (dest) + dest * (1)
			//		= dest + source*dest
			color_shift++; // increase effect 2x because it's usually too pale otherwise (so we don't have to render it twice)
			//	  C = dest + (2*source)*dest
			break;
		case DRAW_ADDSHARP: // similar to DRAW_ADD but colors show more sharply (increases contrast)
			srcBlend = D3DBLEND_SRCCOLOR; // Cs = source
			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (source) + dest * (1)
			//		= dest + source*source
			color_shift++; // increase effect 2x to compensate for brightness lost in contrast change
			//	  C = dest + (2*source)*(2*source)
			break;
		case DRAW_ADDSOFT: // similar to DRAW_ADD but adds less to brighter areas (prevents overexposure)
			srcBlend = D3DBLEND_INVDESTCOLOR; // Cs = 1-dest
			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (1-dest) + dest * (1)
			//		= dest + source*(1-dest)
			//		= dest + source - source*dest
			break;

	}

	if(lastSrcBlend != srcBlend)
	{
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,srcBlend);
		lastSrcBlend = srcBlend;
	}
	if(lastDestBlend != destBlend)
	{
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,destBlend);
		lastDestBlend = destBlend;
	}

	if(alphatype == 1) { // if masked
		device->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE,TRUE); // turn on color keying
	}

	if(draw_state->flags & DRAWOP_INVERTED) { // if inverting, set to use the complement of the source texture
		device->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE|D3DTA_COMPLEMENT);
	}

	if(draw_state->flags & DRAWOP_NOCULL)
		d3d_device->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_NONE);

	D3DTEXTUREOP shade_combine_effect = draw_state->color_effect;

	if(color_shift) {
		if(shade_combine_effect == D3DTOP_MODULATE) {
			if(color_shift == 1)
				shade_combine_effect = D3DTOP_MODULATE2X;
			else
				shade_combine_effect = D3DTOP_MODULATE4X;
		}
		else if(shade_combine_effect == D3DTOP_ADDSIGNED)
			shade_combine_effect = D3DTOP_ADDSIGNED2X;
	}

	if(lastShadeCombineEffect != shade_combine_effect)
	{
		device->SetTextureStageState(0,D3DTSS_COLOROP, shade_combine_effect);
		lastShadeCombineEffect = shade_combine_effect;
	}

	if(draw_state->flags & DRAWOP_NOFILTER) {
		device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_POINT);
		device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTFG_POINT);
	}

	if(draw_state->flags & DRAWOP_NODITHER)
		device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, FALSE);

	if(draw_state->flags & DRAWOP_NOBLENDALPHA && draw_state->flags & DRAW_PLAIN)
		device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);

	HRESULT status = D3D_OK;

	bool new_batch = false;

	if(draw_render_batch_device != device) {
		if(draw_render_batch_open == true)
			status = EndRenderBatch();
		draw_render_batch_device = device;
		new_batch = true;
	}

	if(new_batch || image != last_texture_surface) { // note: image NULL means blank/untextured poly
		last_texture_surface = image;
		status = device->SetTexture(0,image); // set the texture image (or nothing if NULL) to draw
	}

	if(draw_render_batch_open == false) {
		if(SUCCEEDED(status)) {
			status = device->BeginScene();
			if(SUCCEEDED(status))
				draw_render_batch_open = true;
		}
	}

	if(SUCCEEDED(status)) {
		D3DTLVERTEX quad[4];
		float near_plane = draw_state->cam->GetNear();
		float far_plane = draw_state->cam->GetFar();
		float plane_sep = far_plane - near_plane;
		for(int i = 0 ; i < 4 ; i++) {
			quad[i].sx = x[i] - 0.5f;
			quad[i].sy = y[i] - 0.5f;

			if(draw_sort_mode & SCENESORTS)
				if(win_alternator)
					quad[i].sz = 1.0f - (z[i]-near_plane)/(plane_sep)/2.0f; // for signed z-buffer
				else
					quad[i].sz = (z[i]-near_plane)/(plane_sep)/2.0f; // for signed z-buffer
			else
				quad[i].sz = (z[i]-near_plane)/(plane_sep); // for normal z-buffer

			quad[i].rhw = (draw_state->warp_mode!=0) ? draw_state->warp[i] : 1.0f;
			quad[i].color = RGBA_MAKE(draw_state->colors.c[i][0], draw_state->colors.c[i][1], draw_state->colors.c[i][2], draw_state->colors.c[i][3]);
			quad[i].tu = (i==0 || i==2) ? (leftMap) : (rightMap);
			quad[i].tv = (i==0 || i==1) ? (topMap) : (bottomMap);
		}

		status = device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_TLVERTEX, quad,4,0);

		// rest of DRAW_BURN workaround is implemented here, unless DRAW_SUBTRACT workaround also needed:
		if(work_burn && win_subtract_workaround != 2 && (win_subtract_workaround != 1 || alphatype == 2)) {
			srcBlend = D3DBLEND_ZERO; // Cs = 0, or
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_TLVERTEX, quad,4,0); // draw again with DRAW_SUBTRACT:
		}
	}

	// reset DirectX renderstates for next time:

	if(draw_state->flags & DRAWOP_NOBLENDALPHA && draw_state->flags & DRAW_PLAIN)
		device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);

	if(draw_state->flags & DRAWOP_NOFILTER) {
		device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_LINEAR);
		device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTFG_LINEAR);
	}

	if(draw_state->flags & DRAWOP_NODITHER)
		device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE);

	if(alphatype == 1) { // if masked
		device->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE,FALSE);
	}

	if(draw_state->flags & DRAWOP_INVERTED) {
		device->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
	}

	if(draw_state->flags & DRAWOP_NOCULL)
		d3d_device->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CCW);

	// rest of DRAW_BURN workaround is implemented here if the DRAW_SUBTRACT workaround is also needed:
	if(work_burn && !(win_subtract_workaround != 2 && (win_subtract_workaround != 1 || alphatype == 2))) {
		draw_state->flags ^= DRAW_SUBTRACT | DRAW_INVERT; // switch from inversion to subtraction
		draw_state->colors /= 2.0f;
		InternalDrawImageD3D(device, image, src_rect, src_bounds, x, y, z, surf_width, surf_height, alphatype);
		draw_state->colors *= 2.0f;
	}

	if(work_inten) draw_state->colors *= 2.0f; // reverse color change from DRAW_INTENSIFY workaround

	// report our success or failure to do the actual drawing:

	if(FAILED(status)) {
		Direct3DError(status);
		DirectDrawError(status);
		return 0;
	} else {
		return 1;
	}
}

// the simplest image drawing function -- using offscreen surfaces, usually very fast
// includes an option for masked drawing too, just input nomask=0
int WindowsDX::InternalDrawImageDD (LPDIRECTDRAWSURFACE7 p_sourceSurface, LPDIRECTDRAWSURFACE7 p_destSurface, RECT & src_rect, RECT & dst_rect, int nomask)
{
	// this function assumes:
	// both source and destination have already been UNlocked (if not, will refuse to draw anything)
	// the destination has a valid clipper attached to it (if not, won't clip and might crash)
	// (if the source and destination rectangles differ, Blt() will automatically scale the image,
	//	using hardware acceleration which will probably instantly scale it with bicubic filtering or something)

	int options = DDBLT_WAIT;
	if(!nomask) options |= DDBLT_KEYSRC;

	HRESULT status = p_destSurface->Blt(&dst_rect, p_sourceSurface,
										&src_rect,
										options,
										NULL);
	if(FAILED(status)) {
		DirectDrawError(status);
		return 0;
	}
	return 1;
}


RECT WindowsDX::ConvertRECT(RECT & rect, ImageX * img)
{
	// somewhat kludgy way of getting a source rectangle:
	RECT rect2;
	if(rect.right != DEFAULT) {
		memcpy(&rect2,&rect,sizeof(RECT));

		if(img != VIEWPORT && img != NULL) {
			if(img->m_xmult != 1.0f) rect2.left = (LONG)((float)rect2.left * img->m_xmult), rect2.right = (LONG)((float)rect2.right * img->m_xmult);
			if(img->m_ymult != 1.0f) rect2.top = (LONG)((float)rect2.top * img->m_ymult), rect2.bottom = (LONG)((float)rect2.bottom * img->m_ymult);
		}
	} else {
		if(img == VIEWPORT) {
			rect2.left = win_viewx1;
			rect2.top = win_viewy1;
			rect2.right = win_viewx2;
			rect2.bottom = win_viewy2;
		} else if(img == NULL) {
			rect2.left = 0;
			rect2.top = 0;
			rect2.right = 1;
			rect2.bottom = 1;
		} else {
			rect2.left = 0;
			rect2.top = 0;
			rect2.right = img->m_xres;
			rect2.bottom = img->m_yres;
			if(img->m_xmult != 1.0f) rect2.right = (LONG)((float)rect2.right * img->m_xmult);
			if(img->m_ymult != 1.0f) rect2.bottom = (LONG)((float)rect2.bottom * img->m_ymult);
		}
	}
	return rect2;
}

bool WindowsDX::CopyToClipboard(ImageX * img)
{
	LPDIRECTDRAWSURFACE7 p_src_surface;

	if(img == VIEWPORT) {
		p_src_surface = dd_back_buf;
	} else {
		if(img == NULL || img->status != IMG_OK || !img->m_hassurface || img->m_surface == NULL)
			return false;
		p_src_surface = img->m_surface;
	}

	return SurfaceToClipboard(dd_back_buf);
}

bool WindowsDX::SurfaceToClipboard(LPDIRECTDRAWSURFACE7 surface)
{
	bool ok = true;

	HDC hdc;
	DDSURFACEDESC2 ddsd;
	surface->GetDC (&hdc);
	HDC hdc2 = CreateCompatibleDC (hdc); 
	ZeroMemory (&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	surface->GetSurfaceDesc (&ddsd);
	HBITMAP bitmap = CreateCompatibleBitmap (hdc, ddsd.dwWidth, ddsd.dwHeight);
	HBITMAP oldBitmap = (HBITMAP) SelectObject (hdc2, bitmap);
	BitBlt (hdc2, 0, 0, ddsd.dwWidth, ddsd.dwHeight, hdc, 0, 0, SRCCOPY);

	if (OpenClipboard (win_hwnd)) { 
		EmptyClipboard();
		SetClipboardData (CF_BITMAP,bitmap);
		CloseClipboard(); 
	} else
		ok = false;
	
	SelectObject (hdc2,oldBitmap);
	DeleteDC (hdc2);
	surface->ReleaseDC (hdc);
	return ok;
}
